/* bernoulli Attractor PD External */
/* Copyright taken from Willi-Hans Steeb: Chaos and Fractals, 2006 */
/* This program is distributed under the params of the GNU Public License */

///////////////////////////////////////////////////////////////////////////////////
/* This file is part of Chaos PD Externals.                                      */
/*                                                                               */
/* Chaos PD Externals are free software; you can redistribute them and/or modify */
/* them under the terms of the GNU General Public License as published by        */
/* the Free Software Foundation; either version 2 of the License, or             */
/* (at your option) any later version.                                           */
/*                                                                               */
/* Chaos PD Externals are distributed in the hope that they will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 */
/* GNU General Public License for more details.                                  */
/*                                                                               */
/* You should have received a copy of the GNU General Public License             */
/* along with the Chaos PD Externals; if not, write to the Free Software         */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     */
///////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "m_pd.h"



#define M_x 0

#define LY_ITERATIONS 50000
#define LY_ABERATION 10e-15
#define LY_COUNT 3

#define M_param_count 0
#define M_var_count 1
#define M_search_count 0
#define M_limits_count 0
#define M_failure_limit 1000

static char *version = "bernoulli v0.05, by taken from Willi-Hans Steeb: Chaos and Fractals, 2006";

t_class *bernoulli_class;

typedef struct bernoulli_struct {
	t_object x_obj;

	double vars[M_var_count];
	double vars_init[M_var_count];
} bernoulli_struct;

static void calc(bernoulli_struct *bernoulli, double *vars) {
	double x_0=vars[M_x];
	if(x_0 <=0)x_0=.00001;
	x_0 =(x_0<.5)?2*x_0:2*x_0-1;
	vars[M_x] = x_0;
} // end calc

static void calculate(bernoulli_struct *bernoulli) {
	calc(bernoulli, bernoulli -> vars);
	outlet_float(bernoulli -> x_obj.ob_outlet, bernoulli -> vars[M_x]);
} // end calculate

static void reset(bernoulli_struct *bernoulli, t_symbol *s, int argc, t_atom *argv) {
	if (argc == M_var_count) {
		bernoulli -> vars[M_x] = (double) atom_getfloatarg(M_x, argc, argv);
	} else {
		bernoulli -> vars[M_x] = bernoulli -> vars_init[M_x];
	} // end if
} // end reset

void *bernoulli_new(t_symbol *s, int argc, t_atom *argv) {
	bernoulli_struct *bernoulli = (bernoulli_struct *) pd_new(bernoulli_class);
	if (bernoulli != NULL) {
		outlet_new(&bernoulli -> x_obj, &s_float);
		if (argc == M_param_count + M_var_count) {
			bernoulli -> vars_init[M_x] = bernoulli -> vars[M_x] = (double) atom_getfloatarg(0, argc, argv);
		} else {
			if (argc != 0 && argc != M_param_count + M_var_count) {
				post("Incorrect number of arguments for bernoulli fractal. Expecting 1 arguments.");
			}
			bernoulli -> vars_init[M_x] = 0.1;
		}
	}
	return (void *)bernoulli;
}

void bernoulli_setup(void) {
	post(version);
	bernoulli_class = class_new(gensym("bernoulli"), (t_newmethod) bernoulli_new, 0, sizeof(bernoulli_struct), 0, A_GIMME, 0);
	class_addbang(bernoulli_class, (t_method) calculate);
	class_addmethod(bernoulli_class, (t_method) reset, gensym("reset"), A_GIMME, 0);
	class_sethelpsymbol(bernoulli_class, gensym("bernoulli-help.pd"));
}

